

-- --------------------------------------------------------------------------
--
-- Function to fill PERSON_SESSION_ITEM (called by function F_FILL_PERSON_SESSION)
--
-- --------------------------------------------------------------------------
create or replace
FUNCTION F_FILL_PERSON_SESSION_ITEM
(
  P_PERSON_FK IN number,
  P_PERSON_SESSION_FK IN number,
  PRIVILEGE_MANAGER IN number default 1,   -- Attention, MANAGER != admin
  PRIVILEGE_RECORDER IN number default 2,  -- READ/WRITE
  PRIVILEGE_VALIDATOR IN number default 4, -- READ/WRITE/VALIDATE
  OBSERVED_LANDING_OBJ_TYPE IN VARCHAR2 default 'OBSERVED_LANDING',
  OBSERVED_LOCATION_OBJ_TYPE IN VARCHAR2 default 'OBSERVED_LOCATION',
  PROJECT_TEAM_MEMBER_PROFILE IN number default 3,
  YEAR_SPAN_FOR_PTM IN number default 15,  -- NB years of historical data, for project members
  YEAR_SPAN_DEFAULT IN number default 3    -- NB years of historical data, for other users
)
RETURN BOOLEAN
AS
--$ ********************************************************************
--$
--$  MOD : F_FILL_PERSON_SESSION_ITEM
--$  ROL : Compute user access rights, on data without VESSEL_FK.
--$        This function is called by F_FILL_PERSON_SESSION
--$  param :
--$    - P_PERSON_FK: the person identifier
--$    - P_PERSON_SESSION_FK: the person session identifier
--$
--$  return : return a boolean, that say if any row  has been updated
--$
--$ History :
--$  18/03/15 BL Creation (need for Allegro-ObsDeb - mantis #24182)
--$  02/04/15 BL Remove unused with-queries (need for Oracle 10g compatibility)
--$  26/10/15 BL Transform into function, that return if any row  has been updated
--$
--$ ********************************************************************
  session_id number(10);
  CURSOR cur is

      -- Utilisation de sous-requêtes WITH pour plus de clarté

      WITH
      --
      -- AcquisitionLevel des programmes, qui doivent correspondre aux codes des object_type des droits Vessel_Person_Session
      -- via Program > Strategy > PMFM_Strategy > AcquisitionLevel
      --
            program_object_types as
                (select distinct
                  p.code,
                  s.name,
                  -- Make sure 'OBSERVED_LANDING' replaced by 'OBSERVED_LOCATION' (to avoid the creation of new row in PmfmStrategy)
                  decode(pmfm_s.acquisition_level_fk, OBSERVED_LANDING_OBJ_TYPE, OBSERVED_LOCATION_OBJ_TYPE, pmfm_s.acquisition_level_fk) as acquisition_level_fk
                from program p
                  -- jointures internes => les programmes dont l'acquisition_level n'est pas défini seront ignorés
                  inner join strategy s
                    on p.code = s.program_fk
                  inner join pmfm_strategy pmfm_s
                    on pmfm_s.strategy_fk = s.id
                 where pmfm_s.acquisition_level_fk in (
                      OBSERVED_LANDING_OBJ_TYPE,
                      OBSERVED_LOCATION_OBJ_TYPE
                    )
                ),
      --
      -- champs communs (noyau des droits)
      --
      --
      -- droits des personnes
      -- Note sur les superviseurs: un superviseur peut avoir des droits de consultation (viewer) ou validation (validator)
      -- sur les données accessibles à la personne supervisée.
      -- Dans le cas d'un droit de supervision, on remplace donc l'id du droit et le lieu concerné par ceux de la personne supervisée (REFERENCE_PERSON_FK)
      --
            person_program_privilege as
                (select
                  case
                    when p2p.reference_person_fk is not null then ref_p2p.id
                    when p2p.reference_person_fk is null then p2p.id
                  end as id,
                  p2p.person_fk as person,
                  p2p.program_fk,
                  case
                    when p2p.reference_person_fk is not null then ref_p2p.location_fk
                    when p2p.reference_person_fk is null then p2p.location_fk
                  end as location_fk,
                  1 as read_data,
                  case
                    when p2p.program_privilege_fk = PRIVILEGE_VALIDATOR or p2p.program_privilege_fk = PRIVILEGE_RECORDER then 1
                    when p2p.program_privilege_fk <> PRIVILEGE_VALIDATOR and p2p.program_privilege_fk <> PRIVILEGE_RECORDER then 0
                  end as write_data,
                  case
                    when p2p.program_privilege_fk = PRIVILEGE_VALIDATOR then 1
                    when p2p.program_privilege_fk <> PRIVILEGE_VALIDATOR then 0
                  end as validate_data,
                  pot.acquisition_level_fk as object_type_fk
                from program2person p2p
                  inner join program_object_types pot -- object_type_fk n'est pas nullable (relation *-1) => inner join
                    on p2p.program_fk = pot.code
                  left join program2person ref_p2p
                    on p2p.reference_person_fk = ref_p2p.person_fk
                      and p2p.program_fk = ref_p2p.program_fk
                      and ref_p2p.reference_person_fk is null -- to prevent chaining supervisions
                      and ref_p2p.program_privilege_fk <> PRIVILEGE_MANAGER -- managers (superviseurs) do not actually hold rights
                where p2p.person_fk = p_person_fk
                      and p2p.program_privilege_fk <> PRIVILEGE_MANAGER
                      -- avoid rows where reference_person_fk is defined, but no privilege (ref_p2p) exist for this person
                      and (p2p.reference_person_fk is null
                          or ref_p2p.id is not null)
                ),
      --
      -- droits des services (donne la liste des personnes des services concernés)
      --
            department_program_privilege as
                (select p2d.id,
                  p.id as person,
                  p2d.program_fk,
                  p2d.location_fk,
                  1 as read_data,
                  case
                    when p2d.program_privilege_fk = PRIVILEGE_VALIDATOR or p2d.program_privilege_fk = PRIVILEGE_RECORDER then 1
                    when p2d.program_privilege_fk <> PRIVILEGE_VALIDATOR and p2d.program_privilege_fk <> PRIVILEGE_RECORDER then 0
                  end as write_data,
                  case
                    when p2d.program_privilege_fk = PRIVILEGE_VALIDATOR then 1
                    when p2d.program_privilege_fk <> PRIVILEGE_VALIDATOR then 0
                  end as validate_data,
                  pot.acquisition_level_fk as object_type_fk
                from program2department p2d
                  inner join program_object_types pot -- object_type_fk n'est pas nullable (relation *-1) => inner join
                    on p2d.program_fk = pot.code
                  inner join person p
                    on p.department_fk = p2d.department_fk
                where p.id = p_person_fk
                      and p2d.program_privilege_fk <> PRIVILEGE_MANAGER
                ),
      --
      -- Exceptions navires (inclusions et exclusions)
      -- Cette (sous-)requête renvoie les droits qui comportent au moins une exception d'inclusion ainsi que leur nombre
            count_inclusions as
              (select ex_in.program2person_fk as p2p_id,
                    count(ex_in.program2person_fk) as inc_nb
                from program2person_exception ex_in
                where ex_in.is_exclude = 0
                group by ex_in.program2person_fk
              ),
      --
      -- Requêtes principales :
      --  1ère partie: Observations récupérées par le lieu (non nul)
      --
            privileges_by_location as
             (SELECT
                ol.program_fk,
                ppp.read_data,
                ppp.write_data,
                ppp.validate_data,
                ppp.object_type_fk,
                ol.id as object_id,
                ol.start_date_time,
                ol.end_date_time
              FROM
                observed_location ol
                INNER JOIN location_hierarchy lh
                  ON lh.child_location_fk=ol.location_fk
                INNER JOIN person_program_privilege ppp
                  ON ppp.location_fk = lh.parent_location_fk
                LEFT OUTER JOIN landing l
                  ON l.observed_location_fk=ol.id
              WHERE l.id IS NULL
                AND ppp.location_fk IS NOT NULL
                AND ppp.program_fk = ol.program_fk
                AND ppp.object_type_fk = 'OBSERVED_LOCATION'
            ),
      --
      --  2ème partie: Observations when user has right on all locations (location should be null and have no vessel includes)
      --
            privileges_no_location as
            (SELECT ppp.program_fk,
                  max(ppp.read_data) as read_data,
                  max(ppp.write_data) as write_data,
                  max(ppp.validate_data) as validate_data,
                  ppp.object_type_fk,
                  ol.id as object_id,
                  max(ol.start_date_time) as start_date_time,
                  max(ol.end_date_time) as end_date_time
            FROM
                observed_location ol
                LEFT OUTER JOIN landing l
                  ON l.observed_location_fk=ol.id,
                person_program_privilege ppp
                LEFT OUTER JOIN count_inclusions ci
                  ON ci.p2p_id = ppp.id
            WHERE
              l.id is null -- only observation without vessel
              AND ppp.location_fk IS NULL -- seuls les privièges sans aucun lieu défini sont pris en compte
              AND ppp.program_fk = ol.program_fk
              AND ppp.object_type_fk = 'OBSERVED_LOCATION'
              AND (ci.inc_nb IS NULL OR ci.inc_nb = 0)
            GROUP BY ppp.program_fk,
                    ppp.object_type_fk,
                    ol.id
            ),
      --
      --  3ème partie: privilèges hérités du service/société (Program2Department)
      --  ATTENTION!! les privilèges accordés par le department ignorent les exceptions navires
      --  Les privilèges sans restriction de lieux (location_fk = null) ne sont PAS pris en compte
      --
            privileges_by_department as
            (SELECT dpp.program_fk,
                    dpp.read_data,
                    dpp.write_data,
                    dpp.validate_data,
                    dpp.object_type_fk,
                    ol.id as object_id,
                    ol.start_date_time,
                    ol.end_date_time
              FROM
                observed_location ol
                INNER JOIN location_hierarchy lh
                  ON lh.child_location_fk=ol.location_fk
                INNER JOIN department_program_privilege dpp
                  ON dpp.location_fk = lh.parent_location_fk
                LEFT OUTER JOIN landing l
                  ON l.observed_location_fk=ol.id
              WHERE
                l.id is null -- only observation without vessel
                AND dpp.program_fk=ol.program_fk
                AND dpp.object_type_fk = 'OBSERVED_LOCATION'
            ),

      --
      -- requête principale:
      -- Architecture: UNION quadruple
      -- - Lieu défini: les observations sans navires présent sur ce lieu (en prenant en compte la hiérarchie)
      -- - Lieu indéfini ne comportant pas d'exceptions d'inclusion (= tous les lieux = toutes les observations)
      -- - Les droits hérités du service ou de la société de la personne
      -- Note: utilisation d'UNION ALL au lieu d'UNION pour gain de performance (évite le 'distinct')
      --
            privileges_union as
            (SELECT * FROM privileges_by_location
            UNION ALL
            SELECT * FROM privileges_no_location
            UNION ALL
            SELECT * FROM privileges_by_department
            )

      SELECT session_id,
            pu.program_fk,
            max(pu.read_data) as read_data,
            max(pu.write_data) as write_data,
            max(pu.validate_data) as validate_data,
            pu.object_type_fk,
            pu.object_id
      FROM privileges_union pu
        -- ignore vessels registered before 10 years (project team members, ) or 3 UserProfileId 3
        LEFT JOIN PERSON2USER_PROFIL up
          ON up.person_fk = p_person_fk
          AND up.user_profil_fk = PROJECT_TEAM_MEMBER_PROFILE
      WHERE
        pu.start_date_time >= to_date('01-01-'||(extract(year from sysdate)-decode(up.user_profil_fk, PROJECT_TEAM_MEMBER_PROFILE, YEAR_SPAN_FOR_PTM, YEAR_SPAN_DEFAULT)), 'DD-MM-YYYY')
          OR (pu.end_date_time is not null
              AND pu.end_date_time > to_date('01-01-'||(extract(year from sysdate)-decode(up.user_profil_fk, PROJECT_TEAM_MEMBER_PROFILE, YEAR_SPAN_FOR_PTM, YEAR_SPAN_DEFAULT)), 'DD-MM-YYYY')
              AND pu.start_date_time <= to_date('01-01-'||(extract(year from sysdate)-decode(up.user_profil_fk, PROJECT_TEAM_MEMBER_PROFILE, YEAR_SPAN_FOR_PTM, YEAR_SPAN_DEFAULT)), 'DD-MM-YYYY')
             )
      GROUP BY pu.program_fk,
            pu.object_type_fk,
            pu.object_id
      ORDER BY program_fk;

  row_to_check  person_session_item%rowtype;
  TYPE          t_program_by_psi_id is table of VARCHAR2(40) index by PLS_INTEGER;
  program_by_psi_id   t_program_by_psi_id;
  psi_id        person_session_item.id%type;
  deb           integer; -- indice de parcour des programmes
  fin           integer;
  has_update    boolean := FALSE; -- Marqueur pour savoir s'il y des mise à jour effectuées

  procedure p (rec in cur%rowtype, program in varchar2)
  is
     recordToInsert cur%rowtype;
  BEGIN

    -- insertion en table temporaire pour contrôle des suppression
    recordToInsert := rec;
    recordToInsert.program_fk := program;
    BEGIN
      insert into temp_person_session_item (PERSON_SESSION_FK,
                READ_DATA,
                WRITE_DATA,
                VALIDATE_DATA,
                PROGRAM_FK,
                OBJECT_TYPE_FK,
                OBJECT_ID)
      values (recordToInsert.session_id,
        recordToInsert.read_data,
        recordToInsert.write_data,
        recordToInsert.validate_data,
        recordToInsert.program_fk,
        recordToInsert.object_type_fk,
        recordToInsert.object_id);

    EXCEPTION
      WHEN DUP_VAL_ON_INDEX THEN
        DBMS_OUTPUT.PUT_LINE('Ignore exception duplicate index on (PERSON_SESSION_FK, PROGRAM, OBJECT_TYPE, OBJECT_ID):'
        || ' - ' || session_id
        || ' - ' ||  RECORDTOINSERT.PROGRAM_FK
        || ' - ' ||  RECORDTOINSERT.OBJECT_TYPE_FK
        || ' - ' ||  RECORDTOINSERT.OBJECT_ID);
        NULL;
    END;

    select *
    into row_to_check
    from person_session_item
    where person_session_fk = session_id
    and program_fk = program
    and object_type_fk = rec.object_type_fk
    and object_id = rec.object_id;

    -- Contrôle si l'enregistrement en base doit être mis à jour
    if rec.READ_DATA != row_to_check.READ_DATA
    or rec.WRITE_DATA != row_to_check.WRITE_DATA
    or rec.VALIDATE_DATA != row_to_check.VALIDATE_DATA then
      -- Mise à jour d'un droit existant
      update person_session_item
      set READ_DATA = rec.READ_DATA,
          WRITE_DATA = rec.WRITE_DATA,
          VALIDATE_DATA = rec.VALIDATE_DATA,
          update_date = sysdate
      where person_session_fk = session_id
      and program_fk = program
      and object_type_fk = rec.object_type_fk
      and object_id = rec.object_id;

      has_update := TRUE;
    end if;

  EXCEPTION
    WHEN NO_DATA_FOUND THEN
      -- Insertion d'un nouveau droit
      insert into person_session_item(ID,
                          OBJECT_TYPE_FK,
                          OBJECT_ID,
                          PROGRAM_FK,
                          PERSON_SESSION_FK,
                          READ_DATA,
                          WRITE_DATA,
                          VALIDATE_DATA,
                          UPDATE_DATE)
      values (PERSON_SESSION_ITEM_SEQ.nextval,
                          rec.OBJECT_TYPE_FK,
                          rec.OBJECT_ID,
                          program,
                          session_id,
                          rec.READ_DATA,
                          rec.WRITE_DATA,
                          rec.VALIDATE_DATA,
                          systimestamp);

      has_update := TRUE;
  END p;

BEGIN
  -- Retrieve user sessions id and the previous fill duration
  session_id := p_person_session_fk;

  for rec in cur loop
    -- Insérer autant de ligne dans person_session_vessel qu'il y a de programmes
    -- retournés par le curseur
    deb := 1;
    fin := instr(rec.program_fk,',');
    while (fin != 0) loop
      p(rec, substr(rec.program_fk, deb,fin-deb));
      deb := fin + 1;
      fin := instr(rec.program_fk, ',', deb);
    end loop;
    p(rec, substr(rec.program_fk, deb));
  end loop;

  -- Contrôle des suppressions
  for rec in (select p.*
              from person_session_item p
              left join temp_person_session_item t on (p.person_session_fk = t.person_session_fk
                                                    and p.program_fk = t.program_fk
                                                    and p.object_type_fk = t.object_type_fk
                                                    and p.object_id = t.object_id)
              where p.person_session_fk = session_id
              and t.person_session_fk is null
            ) loop

    -- Store of deleted item
    program_by_psi_id(rec.id) := rec.program_fk;

    delete person_session_item p
    where p.person_session_fk = rec.person_session_fk
    and p.program_fk = rec.program_fk
    and p.object_type_fk = rec.object_type_fk
    and p.object_id = rec.object_id;
  end loop;

  -- Insertion des lignes dans deleted_item_history
  psi_id := program_by_psi_id.first;
  while psi_id is not null loop
    insert into deleted_item_history (id, object_id, recorder_person_fk, object_type_fk, update_date, program_fk)
    values (deleted_item_history_seq.nextval, psi_id, p_person_fk, 'PERSON_SESSION_ITEM', sysdate, program_by_psi_id(psi_id));
    psi_id := program_by_psi_id.next(psi_id);

    has_update:= TRUE;
  end loop;

  RETURN has_update;

END F_FILL_PERSON_SESSION_ITEM;
/


-- --------------------------------------------------------------------------
--
-- Function to fill PERSON_SESSION and PERSON_SESSION_VESSEL tables
--
-- --------------------------------------------------------------------------
create or replace
FUNCTION F_FILL_PERSON_SESSION
(
  P_PERSON_FK IN number,
  P_SKIP_LONG_DURATION IN NUMBER DEFAULT 1
)
RETURN SYS_REFCURSOR AS
--$ ********************************************************************
--$
--$  MOD : F_FILL_PERSON_SESSION
--$  ROL : Calcul des droits d'un utilisateur
--$  param :
--$    - P_PERSON_FK: the person identifier
--$    - P_SKIP_LONG_DURATION: (0/1) Set to 1 to disable rights generation for too long execution.
--$                                  Set to 0 to force the rights generation.
--$
--$  Retourner un sys_RefCursor pour être compatible avec Hibernate
--$
--$ History :
--$  09/12/13 LT Refonte de la fonction pour la nouvelle gestion des droits
--$  09/12/13 LT Ajout du type 'navire scientifique'
--$              Passage de la constante YEAR_SPAN_FOR_PTM (profil SIH_SUPER_UTILISATEUR) de 10 à 15 ans
--$  20/12/13 LT Ignorer les doublons de navires (ex: QIM + exception d'inclusion) - mantis #18799
--$  03/10/14 BL Ajout du niveau d'acquisition OBSERVED_LANDING (pour Allegro-ObsDeb - #mantis #22276)
--$  18/03/15 BL Fill PERSON_SESSION_ITEM with OBSERVED_LOCATION (need for Allegro-ObsDeb - mantis #24182)
--$              Fill DELETED_ITEM_HISTORY.PROGRAM_FK
--$  11/06/15 EB Passage du tableau t_table_of_vessels de 12 à 40
--$  26/10/15 BL Update PERSON_SESSION.UPDATE_DATE only if change detected
--$
--$ ********************************************************************
  MAX_DURATION_IN_MS constant number(19) := 30000; -- Max execution time, in millisecond (by default 30s)

  PRIVILEGE_MANAGER constant number(10) := 1; -- Attention, MANAGER != admin
  PRIVILEGE_RECORDER constant number(10) := 2; -- READ/WRITE
  PRIVILEGE_VALIDATOR constant number(10) := 4; -- READ/WRITE/VALIDATE

  VESSEL_DEFAULT_PROGRAM constant varchar2(30) := 'SIH';
  VESSEL_TYPE_PRO constant number(10) := 1;
  VESSEL_TYPE_INFORMAL constant number(10) := 5;
  VESSEL_TYPE_SCIENTIFIC constant number(10) := 8;

  CALENDAR_OBJ_TYPE constant varchar2(30) := 'ACTIVITY_CALENDAR';
  FISHING_TRIP_OBJ_TYPE constant varchar2(30) := 'OBSERVED_FISHING_TRIP';
  PHYS_GEAR_SURVEY_OBJ_TYPE constant varchar2(30) := 'PHYSICAL_GEAR_SURVEY';
  SALE_OBJ_TYPE constant varchar2(30) := 'OBSERVED_SALE';
  SCIENT_CRUISE_OBJ_TYPE constant varchar2(30) := 'SCIENTIFIC_CRUISE';
  OBSERVED_LANDING_OBJ_TYPE constant varchar2(30) := 'OBSERVED_LANDING';
  OBSERVED_LOCATION_OBJ_TYPE constant varchar2(30) := 'OBSERVED_LOCATION';

  PROJECT_TEAM_MEMBER_PROFILE constant number(10) := 3;
  YEAR_SPAN_FOR_PTM constant number(10) := 15;
  YEAR_SPAN_DEFAULT constant number(10) := 3;

  session_id number(10);
  hibernate_result sys_refCursor;
  previous_fill_duration NUMBER(19);
  time_1   NUMBER(19);

  CURSOR cur is

      -- Utilisation de sous-requêtes WITH pour plus de clarté

      WITH
      --
      -- AcquisitionLevel des programmes, qui doivent correspondre aux codes des object_type des droits Vessel_Person_Session
      -- via Program > Strategy > PMFM_Strategy > AcquisitionLevel
      --
            program_object_types as
                (select distinct
                  p.code,
                  s.name,
                  pmfm_s.acquisition_level_fk
                from program p
      -- jointures internes => les programmes dont l'acquisition_level n'est pas défini seront ignorés
                  inner join strategy s
                    on p.code = s.program_fk
                  inner join pmfm_strategy pmfm_s
                    on pmfm_s.strategy_fk = s.id
                 where pmfm_s.acquisition_level_fk in (
                      CALENDAR_OBJ_TYPE,
                      FISHING_TRIP_OBJ_TYPE,
                      PHYS_GEAR_SURVEY_OBJ_TYPE,
                      SALE_OBJ_TYPE,
                      SCIENT_CRUISE_OBJ_TYPE,
                      OBSERVED_LANDING_OBJ_TYPE
                    )
                ),
      --
      -- champs communs (noyau des droits)
      --
      --
      -- droits des personnes
      -- Note sur les superviseurs: un superviseur peut avoir des droits de consultation (viewer) ou validation (validator)
      -- sur les données accessibles à la personne supervisée.
      -- Dans le cas d'un droit de supervision, on remplace donc l'id du droit et le lieu concerné par ceux de la personne supervisée (REFERENCE_PERSON_FK)
      --
            person_program_privilege as
                (select
                  case
                    when p2p.reference_person_fk is not null then ref_p2p.id
                    when p2p.reference_person_fk is null then p2p.id
                  end as id,
                  p2p.person_fk as person,
                  p2p.program_fk,
                  case
                    when p2p.reference_person_fk is not null then ref_p2p.location_fk
                    when p2p.reference_person_fk is null then p2p.location_fk
                  end as location_fk,
                  1 as read_vessel,
                  1 as read_data,
                  case
                    when p2p.program_privilege_fk = PRIVILEGE_VALIDATOR or p2p.program_privilege_fk = PRIVILEGE_RECORDER then 1
                    when p2p.program_privilege_fk <> PRIVILEGE_VALIDATOR and p2p.program_privilege_fk <> PRIVILEGE_RECORDER then 0
                  end as write_data,
                  case
                    when p2p.program_privilege_fk = PRIVILEGE_VALIDATOR then 1
                    when p2p.program_privilege_fk <> PRIVILEGE_VALIDATOR then 0
                  end as validate_data,
                  pot.acquisition_level_fk as object_type_fk
                from program2person p2p
                  inner join program_object_types pot -- object_type_fk n'est pas nullable (relation *-1) => inner join
                    on p2p.program_fk = pot.code
                  left join program2person ref_p2p
                    on p2p.reference_person_fk = ref_p2p.person_fk
                      and p2p.program_fk = ref_p2p.program_fk
                      and ref_p2p.reference_person_fk is null -- to prevent chaining supervisions
                      and ref_p2p.program_privilege_fk <> PRIVILEGE_MANAGER -- managers (superviseurs) do not actually hold rights
                where p2p.person_fk = p_person_fk
                      and p2p.program_privilege_fk <> PRIVILEGE_MANAGER
                      -- avoid rows where reference_person_fk is defined, but no privilege (ref_p2p) exist for this person
                      and (p2p.reference_person_fk is null
                          or ref_p2p.id is not null)
                ),
      --
      -- droits des services (donne la liste des personnes des services concernés)
      --
            department_program_privilege as
                (select p2d.id,
                  p.id as person,
                  p2d.program_fk,
                  p2d.location_fk,
                  1 as read_vessel,
                  1 as read_data,
                  case
                    when p2d.program_privilege_fk = PRIVILEGE_VALIDATOR or p2d.program_privilege_fk = PRIVILEGE_RECORDER then 1
                    when p2d.program_privilege_fk <> PRIVILEGE_VALIDATOR and p2d.program_privilege_fk <> PRIVILEGE_RECORDER then 0
                  end as write_data,
                  case
                    when p2d.program_privilege_fk = PRIVILEGE_VALIDATOR then 1
                    when p2d.program_privilege_fk <> PRIVILEGE_VALIDATOR then 0
                  end as validate_data,
                  pot.acquisition_level_fk as object_type_fk
                from program2department p2d
                  inner join program_object_types pot -- object_type_fk n'est pas nullable (relation *-1) => inner join
                    on p2d.program_fk = pot.code
                  inner join person p
                    on p.department_fk = p2d.department_fk
                where p.id = p_person_fk
                      and p2d.program_privilege_fk <> PRIVILEGE_MANAGER
                ),
      --
      -- Exceptions navires (inclusions et exclusions)
      --
            included_exceptions as
              (select p2pe.program2person_fk as p2p_id,
                  p2pe.vessel_fk
                from program2person_exception p2pe
                where p2pe.is_exclude = 0
              ),
            excluded_exceptions as
              (select p2pe.program2person_fk as p2p_id,
                  p2pe.vessel_fk
                from program2person_exception p2pe
                where p2pe.is_exclude = 1
              ),
      -- Cette (sous-)requête renvoie les droits qui comportent au moins une exception d'inclusion ainsi que leur nombre
            count_inclusions as
              (select ex_in.program2person_fk as p2p_id,
                    count(ex_in.program2person_fk) as inc_nb
                from program2person_exception ex_in
                where ex_in.is_exclude = 0
                group by ex_in.program2person_fk
              ),
      --
      -- Liste de navires (et périodes d'immatriculation) récupérée à partir de la hiérarchie des lieux.
      -- Le quartier d'immatriculation doit être fils du lieu parent spécifié
      --
            vessel_by_parent_location as
              (select vrp.vessel_fk,
                  vrp.start_date_time,
                  vrp.end_date_time,
                  loc_h.parent_location_fk
                from vessel v
                  inner join vessel_registration_period vrp
                    on vrp.vessel_fk = v.code
                  inner join location_hierarchy loc_h
                    on loc_h.child_location_fk = vrp.registration_location_fk
                where
                  -- only SIH program vessels that are professional or informal are considered
                    v.program_fk = VESSEL_DEFAULT_PROGRAM
                    and v.vessel_type_fk in (VESSEL_TYPE_PRO, VESSEL_TYPE_INFORMAL, VESSEL_TYPE_SCIENTIFIC)
              ),
      --
      -- requête principale:
      -- 1ère partie: Navires récupérés par le lieu (non nul) qui ne sont pas exclus par les exceptions
      --
            privileges_by_location as
            (SELECT ppp.program_fk,
                    ppp.read_vessel,
                    ppp.read_data,
                    ppp.write_data,
                    ppp.validate_data,
                    ppp.object_type_fk,
                    vbpl.vessel_fk as vessel,
                    vbpl.start_date_time,
                    vbpl.end_date_time
              FROM person_program_privilege ppp
                INNER JOIN vessel_by_parent_location vbpl
                  ON ppp.location_fk = vbpl.parent_location_fk
              WHERE ppp.location_fk IS NOT NULL
                AND vbpl.vessel_fk IS NOT NULL
                AND vbpl.vessel_fk not in (select ex.vessel_fk
                                            from excluded_exceptions ex
                                            where ex.p2p_id = ppp.id)
            ),
      --
      -- 2ème partie: Navires listés par les exceptions d'inclusion. Que le lieu soit nul ou pas ne change rien
      -- Note: les exceptions d'exclusions ne sont pas prises en compte: si une règle d'inclusion est définie, elle prend le pas sur l'exclusion.
      -- TODO : point ci-dessus à valider
      --
            privileges_by_inc_exceptions as
            (SELECT ppp.program_fk,
                    max(ppp.read_vessel) as read_vessel, -- pour un même tuple personne/programme/navire/objecttype, on prend les droits maximaux
                    max(ppp.read_data) as read_data,
                    max(ppp.write_data) as write_data,
                    max(ppp.validate_data) as validate_data,
                    ppp.object_type_fk,
                    ex.vessel_fk as vessel,
                    min(vrp.start_date_time) as start_date_time,
                    max(nvl(vrp.end_date_time,to_date('01-01-2100','DD-MM-YYYY'))) as end_date_time
              FROM person_program_privilege ppp
                INNER JOIN included_exceptions ex
                  ON ppp.id = ex.p2p_id
                LEFT OUTER JOIN vessel_registration_period vrp
                  ON vrp.vessel_fk = ex.vessel_fk
              GROUP BY ex.vessel_fk,
                      ppp.program_fk,
                      object_type_fk,
                      ppp.location_fk
            ),
      --
      -- 3ème partie: Navires listés par les exceptions d'exclusion dans le cas où aucun lieu n'est spécifié.
      -- Il faut alors prendre tous les navires immatriculés, sauf ceux exclus
      -- ATTENTION!! si pour un ProgramPrivilege, une exception d'inclusion a été définie, elle prend le pas sur les exclusions et il faut alors rejeter ce privilege
      -- (i.e. ne pas prendre tous les navires)
      -- TODO : faire une jointure avec P2P. Et exclure les lignes où P2P.person_fk != PPP.person (= droits supervisés mais TOUS navires => trop). OU ajouter colonne ref_person dans PPP (=> plus besoin de jointure).
      --
            privileges_no_location as
            (SELECT ppp.program_fk,
                  max(ppp.read_vessel) as read_vessel, -- pour un même tuple personne/programme/navire/objecttype, on prend les droits maximaux
                  max(ppp.read_data) as read_data,
                  max(ppp.write_data) as write_data,
                  max(ppp.validate_data) as validate_data,
                  ppp.object_type_fk,
                  vrp.vessel_fk as vessel,
                  min(vrp.start_date_time) as start_date_time,
                  max(nvl(vrp.end_date_time,to_date('01-01-2100','DD-MM-YYYY'))) as end_date_time
            FROM person_program_privilege ppp
                LEFT OUTER JOIN count_inclusions ci
                  ON ci.p2p_id = ppp.id,
                vessel_registration_period vrp
                -- only SIH program vessels that are professional or informal are considered
                inner join vessel v
                  on vrp.vessel_fk = v.code
                    and v.program_fk = VESSEL_DEFAULT_PROGRAM
                    and v.vessel_type_fk in (VESSEL_TYPE_PRO, VESSEL_TYPE_INFORMAL, VESSEL_TYPE_SCIENTIFIC)
            WHERE ppp.location_fk IS NULL -- seuls les privièges sans aucun lieu défini sont pris en compte
              AND (ci.inc_nb IS NULL OR ci.inc_nb = 0)
              AND vrp.vessel_fk not in (select ex.vessel_fk
                                          from excluded_exceptions ex
                                          where ex.p2p_id = ppp.id)
            GROUP BY vrp.vessel_fk,
                    ppp.program_fk,
                    object_type_fk
            ),
      --
      -- 4ème partie: privilèges hérités du servicesociété (Program2Department)
      -- ATTENTION!! les privilèges accordés par le department ignorent les exceptions navires
      -- Les privilèges sans restriction de lieux (location_fk = null) ne sont pas pris en compte
      -- => besoin d'exceptions navires par département ?
      --
            privileges_by_department as
            (SELECT dpp.program_fk,
                    dpp.read_vessel,
                    dpp.read_data,
                    dpp.write_data,
                    dpp.validate_data,
                    dpp.object_type_fk,
                    vbpl.vessel_fk as vessel,
                    vbpl.start_date_time,
                    vbpl.end_date_time
              FROM department_program_privilege dpp
                INNER JOIN vessel_by_parent_location vbpl
                  ON dpp.location_fk = vbpl.parent_location_fk
              WHERE dpp.location_fk IS NOT NULL
                AND vbpl.vessel_fk IS NOT NULL
            ),


      --
      -- requête principale:
      -- Architecture: UNION quadruple
      -- - Lieu défini: les navires immatriculés dans ce lieu (en prenant en compte la hiérarchie)
      -- - Lieu défini ou non: ajout des exceptions d'inclusion
      -- - Lieu indéfini ne comportant pas d'exceptions d'inclusion (= tous les navires sauf ceux exclus)
      -- - Les droits hérités du service ou de la société de la personne
      -- Note: utilisation d'UNION ALL au lieu d'UNION pour gain de performance (évite le 'distinct')
      --
            privileges_union as
            (SELECT * FROM privileges_by_location
            UNION ALL
            SELECT * FROM privileges_by_inc_exceptions
            UNION ALL
            SELECT * FROM privileges_no_location
            UNION ALL
            SELECT * FROM privileges_by_department
            )

      SELECT session_id,
            pu.vessel as vessel_fk,
            max(pu.read_vessel) as read_vessel, -- pour un même triplet programme/navire/objecttype, on prend les droits maximaux
            max(pu.read_data) as read_data,
            max(pu.write_data) as write_data,
            max(pu.validate_data) as validate_data,
            pu.program_fk,
            pu.object_type_fk,
            pu.start_date_time as unique_sdt,
            min(greatest(pu.start_date_time, to_date('01-01-'||(extract(year from sysdate)-decode(up.user_profil_fk, PROJECT_TEAM_MEMBER_PROFILE, YEAR_SPAN_FOR_PTM, YEAR_SPAN_DEFAULT)), 'DD-MM-YYYY'))) as start_date_time,
            max(nvl(pu.end_date_time, to_date('01-01-2100','DD-MM-YYYY'))) as end_date_time
      FROM privileges_union pu
      -- ignore vessels registered before 10 years (project team members, ) or 3 UserProfileId 3
        LEFT JOIN PERSON2USER_PROFIL up
          ON up.person_fk = p_person_fk
          AND up.user_profil_fk = PROJECT_TEAM_MEMBER_PROFILE
      WHERE pu.end_date_time is null
	    OR pu.start_date_time >= to_date('01-01-'||(extract(year from sysdate)-decode(up.user_profil_fk, PROJECT_TEAM_MEMBER_PROFILE, YEAR_SPAN_FOR_PTM, YEAR_SPAN_DEFAULT)), 'DD-MM-YYYY')
            OR (pu.end_date_time > to_date('01-01-'||(extract(year from sysdate)-decode(up.user_profil_fk, PROJECT_TEAM_MEMBER_PROFILE, YEAR_SPAN_FOR_PTM, YEAR_SPAN_DEFAULT)), 'DD-MM-YYYY')
                AND pu.start_date_time <= to_date('01-01-'||(extract(year from sysdate)-decode(up.user_profil_fk, PROJECT_TEAM_MEMBER_PROFILE, YEAR_SPAN_FOR_PTM, YEAR_SPAN_DEFAULT)), 'DD-MM-YYYY')
                )
      GROUP BY pu.vessel,
            pu.program_fk,
            pu.object_type_fk,
            pu.start_date_time
      ORDER BY program_fk
;

  row_to_check person_session_vessel%rowtype;
  TYPE t_table_of_vessels is table of VARCHAR2(40) index by varchar2(40);
  table_of_vessels t_table_of_vessels;
  vessel_to_delete vessel.code%type;
  deb integer; -- indice de parcour des programmes
  fin integer;

  -- Marqueurs pour savoir s'il y des mise à jour effectuées
  has_update    boolean := FALSE;
  has_update_in_items boolean := FALSE;

  procedure p (rec in cur%rowtype, program in varchar2)
  is
     recordToInsert cur%rowtype;
  BEGIN

      -- insertion en table temporaire pour contrôle des suppression
      recordToInsert := rec;
      recordToInsert.program_fk := program;
      BEGIN
	insert into temp_person_session_vessel (PERSON_SESSION_FK,
						VESSEL_FK,
						READ_VESSEL,
						READ_DATA,
						WRITE_DATA,
						VALIDATE_DATA,
						PROGRAM_FK,
						OBJECT_TYPE_FK,
						START_DATE_TIME,
						END_DATE_TIME)
	values (recordToInsert.session_id,
		recordToInsert.vessel_fk,
		recordToInsert.read_vessel,
		recordToInsert.read_data,
		recordToInsert.write_data,
		recordToInsert.validate_data,
		recordToInsert.program_fk,
		recordToInsert.object_type_fk,
		recordToInsert.start_date_time,
		recordToInsert.end_date_time);

      EXCEPTION
	WHEN DUP_VAL_ON_INDEX THEN
	  DBMS_OUTPUT.PUT_LINE('Ignore exception duplicate index on (VESSEL, PROGRAM, OBJECT_TYPE, START_DATE):'
		|| ' - ' || RECORDTOINSERT.VESSEL_FK
		|| ' - ' ||  RECORDTOINSERT.PROGRAM_FK
		|| ' - ' ||  RECORDTOINSERT.OBJECT_TYPE_FK
		|| ' - ' ||  RECORDTOINSERT.START_DATE_TIME);
	  NULL;
  END;

      select *
      into row_to_check
      from person_session_vessel
      where person_session_fk = session_id
      and vessel_fk = rec.vessel_fk
      and program_fk = program
      and object_type_fk = rec.object_type_fk
      and start_date_time = rec.start_date_time;

      -- Contrôle si l'enregistrement en base doit être mis à jour
      if rec.READ_VESSEL != row_to_check.READ_VESSEL
      or rec.READ_DATA != row_to_check.READ_DATA
      or rec.WRITE_DATA != row_to_check.WRITE_DATA
      or rec.VALIDATE_DATA != row_to_check.VALIDATE_DATA
      or rec.END_DATE_TIME != row_to_check.END_DATE_TIME then
        -- Mise à jour d'un droit existant
        update person_session_vessel
        set READ_VESSEL = rec.READ_VESSEL,
            READ_DATA = rec.READ_DATA,
            WRITE_DATA = rec.WRITE_DATA,
            VALIDATE_DATA = rec.VALIDATE_DATA,
            END_DATE_TIME = rec.END_DATE_TIME,
            update_date = sysdate
        where person_session_fk = session_id
        and vessel_fk = rec.vessel_fk
        and program_fk = program
        and object_type_fk = rec.object_type_fk
        and start_date_time = rec.start_date_time;

        has_update := TRUE;
      end if;

    EXCEPTION
      WHEN NO_DATA_FOUND THEN
        -- Insertion d'un nouveau droit
        insert into person_session_vessel( ID, START_DATE_TIME,
                            OBJECT_TYPE_FK,
                            VESSEL_FK,
                            PROGRAM_FK,
                            PERSON_SESSION_FK,
                            READ_VESSEL,
                            READ_DATA,
                            WRITE_DATA,
                            VALIDATE_DATA,
                            END_DATE_TIME,
                            UPDATE_DATE)
        values (PERSON_SESSION_VESSEL_SEQ.nextval, rec.START_DATE_TIME,
                            rec.OBJECT_TYPE_FK,
                            rec.VESSEL_FK,
                            program,
                            session_id,
                            rec.READ_VESSEL,
                            rec.READ_DATA,
                            rec.WRITE_DATA,
                            rec.VALIDATE_DATA,
                            rec.END_DATE_TIME,
                            systimestamp);

        has_update := TRUE;
    END p;

BEGIN
  time_1 := DBMS_UTILITY.GET_TIME;

  -- Retrieve user sessions id and the previous fill duration
  BEGIN
    select id, fill_duration into session_id, previous_fill_duration
    from person_session
    where person_fk = p_person_fk;
  EXCEPTION
    WHEN NO_DATA_FOUND THEN
      select person_session_seq.nextval into session_id
      from dual;
      insert into person_session (id, person_fk, update_date)
      values (session_id, p_person_fk, sysdate);
      previous_fill_duration := 0;
  END;

  -- If control on duration is enable
  -- compute rights only if previous fill duration was under the max limit
  if (P_SKIP_LONG_DURATION = 0
     or previous_fill_duration < MAX_DURATION_IN_MS) then

    DBMS_OUTPUT.PUT_LINE('Start to fill person_session for PERSON_FK=' || p_person_fk);

    for rec in cur loop
      -- Insérer autant de ligne dans person_session_vessel qu'il y a de programmes
      -- retournés par le curseur
      deb := 1;
      fin := instr(rec.program_fk,',');
      while (fin != 0) loop
        p(rec, substr(rec.program_fk, deb,fin-deb));
        deb := fin + 1;
        fin := instr(rec.program_fk, ',', deb);
      end loop;
      p(rec, substr(rec.program_fk, deb));
    end loop;

    -- Contrôle des suppressions
    for rec in (select p.*
                from person_session_vessel p
                left join temp_person_session_vessel t on (p.person_session_fk = t.person_session_fk
                                                      and p.vessel_fk = t.vessel_fk
                                                      and p.program_fk = t.program_fk
                                                      and p.object_type_fk = t.object_type_fk
                                                      and p.start_date_time = t.start_date_time)
                where p.person_session_fk = session_id
                and t.person_session_fk is null) loop

      table_of_vessels(rec.vessel_fk) := rec.program_fk;

      delete person_session_vessel p
      where p.person_session_fk = rec.person_session_fk
      and p.vessel_fk = rec.vessel_fk
      and p.program_fk = rec.program_fk
      and p.object_type_fk = rec.object_type_fk
      and p.start_date_time = rec.start_date_time;
    end loop;

    -- Insertion des lignes dans deleted_item_history
    vessel_to_delete := table_of_vessels.first;
    while vessel_to_delete is not null loop
      insert into deleted_item_history (id, vessel_fk, recorder_person_fk, object_type_fk, update_date, program_fk)
      values (deleted_item_history_seq.nextval, vessel_to_delete, p_person_fk, 'PERSON_SESSION_VESSEL', sysdate, table_of_vessels(vessel_to_delete));
      vessel_to_delete := table_of_vessels.next(vessel_to_delete);

      has_update := TRUE;
    end loop;

    -- Fill table person_session_item (see Allegro-ObsDeb mantis #24182)
    has_update_in_items := F_FILL_PERSON_SESSION_ITEM(
      P_PERSON_FK => P_PERSON_FK,
      P_PERSON_SESSION_FK => session_id,
      PRIVILEGE_MANAGER => PRIVILEGE_MANAGER,
      PRIVILEGE_RECORDER => PRIVILEGE_RECORDER,
      PRIVILEGE_VALIDATOR => PRIVILEGE_VALIDATOR,
      OBSERVED_LANDING_OBJ_TYPE => OBSERVED_LANDING_OBJ_TYPE,
      OBSERVED_LOCATION_OBJ_TYPE => OBSERVED_LOCATION_OBJ_TYPE,
      PROJECT_TEAM_MEMBER_PROFILE => PROJECT_TEAM_MEMBER_PROFILE,
      YEAR_SPAN_FOR_PTM => YEAR_SPAN_FOR_PTM,
      YEAR_SPAN_DEFAULT => YEAR_SPAN_DEFAULT
    );

    -- Update session with the new fill duration
    IF (has_update OR has_update_in_items) THEN
      update person_session
      set fill_duration = (DBMS_UTILITY.GET_TIME - time_1),
      update_date = systimestamp
      where id=session_id;
    END IF;

    DBMS_OUTPUT.PUT_LINE('End to fill person_session, in ' || (DBMS_UTILITY.GET_TIME - time_1) || ' ms.');
  end if;

  -- return (for hibernate compatibility)
  OPEN hibernate_result
  FOR select session_id from dual;
  RETURN hibernate_result;

END F_FILL_PERSON_SESSION;
/
